@@ -1,13 +1,14 @@
 import mimetypes
-from os import path
 import re
 import uuid
 
+from os import path
+from requests import Response
 from typing import BinaryIO, List, Optional
 from urllib.parse import urlparse, urlencode, quote
 
 from toot import App, User, http, CLIENT_NAME, CLIENT_WEBSITE
-from toot.exceptions import AuthenticationError, ConsoleError
+from toot.exceptions import ConsoleError
 from toot.utils import drop_empty_values, str_bool, str_bool_nullable
 
 
@@ -30,26 +31,54 @@ def find_account(app, user, account_name):
             normalized_name = username
 
     response = search(app, user, account_name, type="accounts", resolve=True)
-    for account in response["accounts"]:
+    for account in response.json()["accounts"]:
         if account["acct"].lower() == normalized_name:
             return account
 
     raise ConsoleError("Account not found")
 
 
-def _account_action(app, user, account, action):
+def lookup(app, user, acct):
+    return http.get(app, user, "/api/v1/accounts/lookup", {"acct": acct})
+
+
+def _account_action(app, user, account, action) -> Response:
     url = f"/api/v1/accounts/{account}/{action}"
-    return http.post(app, user, url).json()
+    return http.post(app, user, url)
 
 
-def _status_action(app, user, status_id, action, data=None):
+def _status_action(app, user, status_id, action, data=None) -> Response:
+    status_id = _resolve_status_id(app, user, status_id)
     url = f"/api/v1/statuses/{status_id}/{action}"
-    return http.post(app, user, url, data=data).json()
+    return http.post(app, user, url, data=data)
 
 
-def _tag_action(app, user, tag_name, action):
+def _resolve_status_id(app, user, id_or_url) -> str:
+    """
+    If given an URL instead of status ID, attempt to resolve the status ID.
+
+    TODO: Not 100% sure this is the correct way of doing this, but it seems to
+    work for all test cases I've thrown at it. So leaving it undocumented until
+    we're happy it works.
+    """
+    if re.match(r"^https?://", id_or_url):
+        response = search(app, user, id_or_url, resolve=True, type="statuses")
+        statuses = response.json().get("statuses")
+
+        if not statuses:
+            raise ConsoleError(f"Cannot find status matching URL {id_or_url}")
+
+        if len(statuses) > 1:
+            raise ConsoleError(f"Found multiple statuses mathcing URL {id_or_url}")
+
+        return statuses[0]["id"]
+
+    return id_or_url
+
+
+def _tag_action(app, user, tag_name, action) -> Response:
     url = f"/api/v1/tags/{tag_name}/{action}"
-    return http.post(app, user, url).json()
+    return http.post(app, user, url)
 
 
 def create_app(base_url):
@@ -139,7 +168,7 @@ def fetch_app_token(app):
     return http.anon_post(f"{app.base_url}/oauth/token", json=json).json()
 
 
-def login(app, username, password):
+def login(app: App, username: str, password: str):
     url = app.base_url + '/oauth/token'
 
     data = {
@@ -151,16 +180,10 @@ def login(app, username, password):
         'scope': SCOPES,
     }
 
-    response = http.anon_post(url, data=data, allow_redirects=False)
-
-    # If auth fails, it redirects to the login page
-    if response.is_redirect:
-        raise AuthenticationError()
-
-    return response.json()
+    return http.anon_post(url, data=data).json()
 
 
-def get_browser_login_url(app):
+def get_browser_login_url(app: App) -> str:
     """Returns the URL for manual log in via browser"""
     return "{}/oauth/authorize/?{}".format(app.base_url, urlencode({
         "response_type": "code",
@@ -170,7 +193,7 @@ def get_browser_login_url(app):
     }))
 
 
-def request_access_token(app, authorization_code):
+def request_access_token(app: App, authorization_code: str):
     url = app.base_url + '/oauth/token'
 
     data = {
@@ -188,7 +211,7 @@ def post_status(
     app,
     user,
     status,
-    visibility='public',
+    visibility=None,
     media_ids=None,
     sensitive=False,
     spoiler_text=None,
@@ -200,7 +223,7 @@ def post_status(
     poll_expires_in=None,
     poll_multiple=None,
     poll_hide_totals=None,
-):
+) -> Response:
     """
     Publish a new status.
     https://docs.joinmastodon.org/methods/statuses/#create
@@ -232,7 +255,53 @@ def post_status(
             "hide_totals": poll_hide_totals,
         }
 
-    return http.post(app, user, '/api/v1/statuses', json=data, headers=headers).json()
+    return http.post(app, user, '/api/v1/statuses', json=data, headers=headers)
+
+
+def edit_status(
+    app,
+    user,
+    id,
+    status,
+    visibility='public',
+    media_ids=None,
+    sensitive=False,
+    spoiler_text=None,
+    in_reply_to_id=None,
+    language=None,
+    content_type=None,
+    poll_options=None,
+    poll_expires_in=None,
+    poll_multiple=None,
+    poll_hide_totals=None,
+) -> Response:
+    """
+    Edit an existing status
+    https://docs.joinmastodon.org/methods/statuses/#edit
+    """
+
+    # Strip keys for which value is None
+    # Sending null values doesn't bother Mastodon, but it breaks Pleroma
+    data = drop_empty_values({
+        'status': status,
+        'media_ids': media_ids,
+        'visibility': visibility,
+        'sensitive': sensitive,
+        'in_reply_to_id': in_reply_to_id,
+        'language': language,
+        'content_type': content_type,
+        'spoiler_text': spoiler_text,
+    })
+
+    if poll_options:
+        data["poll"] = {
+            "options": poll_options,
+            "expires_in": poll_expires_in,
+            "multiple": poll_multiple,
+            "hide_totals": poll_hide_totals,
+        }
+
+    return http.put(app, user, f"/api/v1/statuses/{id}", json=data)
 
 
 def fetch_status(app, user, id):
@@ -240,7 +309,16 @@ def fetch_status(app, user, id):
     Fetch a single status
     https://docs.joinmastodon.org/methods/statuses/#get
     """
-    return http.get(app, user, f"/api/v1/statuses/{id}").json()
+    return http.get(app, user, f"/api/v1/statuses/{id}")
+
+
+def fetch_status_source(app, user, id):
+    """
+    Fetch the source (original text) for a single status.
+    This only works on local toots.
+    https://docs.joinmastodon.org/methods/statuses/#source
+    """
+    return http.get(app, user, f"/api/v1/statuses/{id}/source")
 
 
 def scheduled_statuses(app, user):
@@ -295,14 +373,36 @@ def translate(app, user, status_id):
     return _status_action(app, user, status_id, 'translate')
 
 
-def context(app, user, status_id):
+def context(app, user, status_id) -> Response:
     url = f"/api/v1/statuses/{status_id}/context"
-    return http.get(app, user, url).json()
+    return http.get(app, user, url)
 
 
-def reblogged_by(app, user, status_id):
+def reblogged_by(app, user, status_id) -> Response:
     url = f"/api/v1/statuses/{status_id}/reblogged_by"
-    return http.get(app, user, url).json()
+    return http.get(app, user, url)
+
+
+def get_timeline_generator(
+    app: Optional[App],
+    user: Optional[User],
+    account: Optional[str] = None,
+    list_id: Optional[str] = None,
+    tag: Optional[str] = None,
+    local: bool = False,
+    public: bool = False,
+    limit: int = 20,  # TODO
+):
+    if public:
+        return public_timeline_generator(app, user, local=local, limit=limit)
+    elif tag:
+        return tag_timeline_generator(app, user, tag, local=local, limit=limit)
+    elif account:
+        return account_timeline_generator(app, user, account, limit=limit)
+    elif list_id:
+        return timeline_list_generator(app, user, list_id, limit=limit)
+    else:
+        return home_timeline_generator(app, user, limit=limit)
 
 
 def _get_next_path(headers):
@@ -314,27 +414,32 @@ def _get_next_path(headers):
         return "?".join([parsed.path, parsed.query])
 
 
+def _get_next_url(headers) -> Optional[str]:
+    """Given timeline response headers, returns the url to the next batch"""
+    links = headers.get('Link', '')
+    match = re.match('<([^>]+)>; rel="next"', links)
+    if match:
+        return match.group(1)
+
+
 def _timeline_generator(app, user, path, params=None):
+    if params:
+        path += f"?{urlencode(params)}"
+
     while path:
-        response = http.get(app, user, path, params)
+        response = http.get(app, user, path)
         yield response.json()
         path = _get_next_path(response.headers)
 
 
 def _notification_timeline_generator(app, user, path, params=None):
-    while path:
-        response = http.get(app, user, path, params)
-        notification = response.json()
-        yield [n["status"] for n in notification if n["status"]]
-        path = _get_next_path(response.headers)
+    for batch in _timeline_generator(app, user, path, params):
+        yield [n["status"] for n in batch if n.get("status")]
 
 
 def _conversation_timeline_generator(app, user, path, params=None):
-    while path:
-        response = http.get(app, user, path, params)
-        conversation = response.json()
-        yield [c["last_status"] for c in conversation if c["last_status"]]
-        path = _get_next_path(response.headers)
+    for batch in _timeline_generator(app, user, path, params):
+        yield [c["last_status"] for c in batch if c.get("last_status")]
 
 
 def home_timeline_generator(app, user, limit=20):
@@ -361,6 +466,12 @@ def bookmark_timeline_generator(app, user, limit=20):
     return _timeline_generator(app, user, path, params)
 
 
+def favourite_timeline_generator(app, user, limit=20):
+    path = '/api/v1/favourites'
+    params = {'limit': limit}
+    return _timeline_generator(app, user, path, params)
+
+
 def notification_timeline_generator(app, user, limit=20):
     # exclude all but mentions and statuses
     exclude_types = ["follow", "favourite", "reblog", "poll", "follow_request"]
@@ -374,36 +485,44 @@ def conversation_timeline_generator(app, user, limit=20):
     return _conversation_timeline_generator(app, user, path, params)
 
 
-def account_timeline_generator(app: App, user: User, account_name: str, replies=False, reblogs=False, limit=20):
+def account_timeline_generator(app, user, account_name: str, replies=False, reblogs=False, limit=20):
     account = find_account(app, user, account_name)
     path = f"/api/v1/accounts/{account['id']}/statuses"
     params = {"limit": limit, "exclude_replies": not replies, "exclude_reblogs": not reblogs}
     return _timeline_generator(app, user, path, params)
 
 
+def account_timeline_generator_by_id(app, user, account_id: str, replies=False, reblogs=False, limit=20):
+    path = f"/api/v1/accounts/{account_id}/statuses"
+    params = {"limit": limit, "exclude_replies": not replies, "exclude_reblogs": not reblogs}
+    return _timeline_generator(app, user, path, params)
+
+
 def timeline_list_generator(app, user, list_id, limit=20):
     path = f"/api/v1/timelines/list/{list_id}"
     return _timeline_generator(app, user, path, {'limit': limit})
 
 
-def _anon_timeline_generator(instance, path, params=None):
-    while path:
-        url = f"https://{instance}{path}"
-        response = http.anon_get(url, params)
+def _anon_timeline_generator(url, params=None):
+    if params:
+        url += f"?{urlencode(params)}"
+
+    while url:
+        response = http.anon_get(url)
         yield response.json()
-        path = _get_next_path(response.headers)
+        url = _get_next_url(response.headers)
 
 
-def anon_public_timeline_generator(instance, local=False, limit=20):
-    path = '/api/v1/timelines/public'
-    params = {'local': str_bool(local), 'limit': limit}
-    return _anon_timeline_generator(instance, path, params)
+def anon_public_timeline_generator(base_url, local=False, limit=20):
+    params = {"local": str_bool(local), "limit": limit}
+    url = f"{base_url}/api/v1/timelines/public"
+    return _anon_timeline_generator(url, params)
 
 
-def anon_tag_timeline_generator(instance, hashtag, local=False, limit=20):
-    path = f"/api/v1/timelines/tag/{quote(hashtag)}"
-    params = {'local': str_bool(local), 'limit': limit}
-    return _anon_timeline_generator(instance, path, params)
+def anon_tag_timeline_generator(base_url, hashtag, local=False, limit=20):
+    query = {"local": str_bool(local), "limit": limit}
+    url = f"{base_url}/api/v1/timelines/tag/{quote(hashtag)}"
+    return _anon_timeline_generator(url, query)
 
 
 def get_media(app: App, user: User, id: str):
@@ -426,7 +545,7 @@ def upload_media(
         "thumbnail": _add_mime_type(thumbnail)
     })
 
-    return http.post(app, user, "/api/v2/media", data=data, files=files).json()
+    return http.post(app, user, "/api/v2/media", data=data, files=files)
 
 
 def _add_mime_type(file):
@@ -446,16 +565,32 @@ def _add_mime_type(file):
     return (filename, file, mime_type)
 
 
-def search(app, user, query, resolve=False, type=None):
+def search(
+    app,
+    user,
+    query,
+    resolve=False,
+    type=None,
+    offset=None,
+    limit=None,
+    min_id=None,
+    max_id=None,
+):
     """
     Perform a search.
     https://docs.joinmastodon.org/methods/search/#v2
     """
-    return http.get(app, user, "/api/v2/search", {
+    params = drop_empty_values({
         "q": query,
-        "resolve": resolve,
-        "type": type
-    }).json()
+        "resolve": str_bool(resolve),
+        "type": type,
+        "offset": offset,
+        "limit": limit,
+        "min_id": min_id,
+        "max_id": max_id,
+    })
+
+    return http.get(app, user, "/api/v2/search", params)
 
 
 def follow(app, user, account):
@@ -466,11 +601,11 @@ def unfollow(app, user, account):
     return _account_action(app, user, account, 'unfollow')
 
 
-def follow_tag(app, user, tag_name):
+def follow_tag(app, user, tag_name) -> Response:
     return _tag_action(app, user, tag_name, 'follow')
 
 
-def unfollow_tag(app, user, tag_name):
+def unfollow_tag(app, user, tag_name) -> Response:
     return _tag_action(app, user, tag_name, 'unfollow')
 
 
@@ -498,6 +633,43 @@ def followed_tags(app, user):
     return _get_response_list(app, user, path)
 
 
+def featured_tags(app, user):
+    return http.get(app, user, "/api/v1/featured_tags")
+
+
+def feature_tag(app, user, tag: str) -> Response:
+    return http.post(app, user, "/api/v1/featured_tags", data={"name": tag})
+
+
+def unfeature_tag(app, user, tag_id: str) -> Response:
+    return http.delete(app, user, f"/api/v1/featured_tags/{tag_id}")
+
+
+def find_tag(app, user, tag) -> Optional[dict]:
+    """Find a hashtag by tag name or ID"""
+    tag = tag.lstrip("#")
+    results = search(app, user, tag, type="hashtags").json()
+
+    return next(
+        (
+            t for t in results["hashtags"]
+            if t["name"].lower() == tag.lstrip("#").lower() or t["id"] == tag
+        ),
+        None
+    )
+
+
+def find_featured_tag(app, user, tag) -> Optional[dict]:
+    """Find a featured tag by tag name or ID"""
+    return next(
+        (
+            t for t in featured_tags(app, user).json()
+            if t["name"].lower() == tag.lstrip("#").lower() or t["id"] == tag
+        ),
+        None
+    )
+
+
 def whois(app, user, account):
     return http.get(app, user, f'/api/v1/accounts/{account}').json()
 
@@ -537,40 +709,39 @@ def blocked(app, user):
     return _get_response_list(app, user, "/api/v1/blocks")
 
 
-def verify_credentials(app, user):
-    return http.get(app, user, '/api/v1/accounts/verify_credentials').json()
+def verify_credentials(app, user) -> Response:
+    return http.get(app, user, '/api/v1/accounts/verify_credentials')
 
 
-def single_status(app, user, status_id):
-    url = f"/api/v1/statuses/{status_id}"
-    return http.get(app, user, url).json()
-
-
-def get_notifications(app, user, exclude_types=[], limit=20):
-    params = {"exclude_types[]": exclude_types, "limit": limit}
-    return http.get(app, user, '/api/v1/notifications', params).json()
+def get_notifications(app, user, types=[], exclude_types=[], limit=20) -> Response:
+    params = {"types[]": types, "exclude_types[]": exclude_types, "limit": limit}
+    return http.get(app, user, '/api/v1/notifications', params)
 
 
 def clear_notifications(app, user):
     http.post(app, user, '/api/v1/notifications/clear')
 
 
-def get_instance(base_url):
+def get_instance(base_url: str) -> Response:
     url = f"{base_url}/api/v1/instance"
-    return http.anon_get(url).json()
+    return http.anon_get(url)
+
+
+def get_preferences(app, user) -> Response:
+    return http.get(app, user, '/api/v1/preferences')
 
 
 def get_lists(app, user):
-    path = "/api/v1/lists"
-    return _get_response_list(app, user, path)
+    return http.get(app, user, "/api/v1/lists").json()
 
 
-def find_list_id(app, user, title):
-    lists = get_lists(app, user)
-    for list_item in lists:
-        if list_item["title"] == title:
-            return list_item["id"]
-    return None
+def get_poll(app, user, poll_id) -> Response:
+    return http.get(app, user, f"/api/v1/polls/{poll_id}")
+
+
+def vote_poll(app, user, poll_id, choices) -> Response:
+    json = {"choices": choices}
+    return http.post(app, user, f"/api/v1/polls/{poll_id}/votes", json=json)
 
 
 def get_list_accounts(app, user, list_id):
@@ -578,12 +749,12 @@ def get_list_accounts(app, user, list_id):
     return _get_response_list(app, user, path)
 
 
-def create_list(app, user, title, replies_policy):
+def create_list(app, user, title, replies_policy="none"):
     url = "/api/v1/lists"
     json = {'title': title}
     if replies_policy:
         json['replies_policy'] = replies_policy
-    return http.post(app, user, url, json=json).json()
+    return http.post(app, user, url, json=json)
 
 
 def delete_list(app, user, id):
@@ -593,7 +764,7 @@ def delete_list(app, user, id):
 def add_accounts_to_list(app, user, list_id, account_ids):
     url = f"/api/v1/lists/{list_id}/accounts"
     json = {'account_ids': account_ids}
-    return http.post(app, user, url, json=json).json()
+    return http.post(app, user, url, json=json)
 
 
 def remove_accounts_from_list(app, user, list_id, account_ids):
